#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Instrument.py
#	
#	This module is the non-gui class the represents instruments. instruments
#	represent a track of audio that can contain many different sources in sequence.
#	This module handles loading and saving instruments from xml, the gstreamer
#	bits for playing and recording events, audio effects plugins, as well as any 
#	instrument specific functionality like; solo, mute, volume, etc.
#
#-------------------------------------------------------------------------------

from Event import *
from UndoSystem import *
import pygst
pygst.require("0.10")
import gst
import os
import time
from Monitored import *
from Utils import *
import gobject
import gnomevfs
import Globals
import shutil
import AlsaDevices
import gettext
_ = gettext.gettext

#=========================================================================	

class Instrument(Monitored):
	
	#_____________________________________________________________________
	
	def __init__(self, project, name, type, pixbuf, id=None):
		Monitored.__init__(self)
		
		self.project = project
		
		self.path = ""					# The 'audio' directory for this instrument
		self.events = []				# List of events attached to this instrument
		self.graveyard = []				# List of events that have been deleted (kept for undo)
		self.name = name				# Name of this instrument
		self.pixbuf = pixbuf			# The icon pixbuf resource
		self.isArmed = False			# True if the instrument is armed for recording
		self.isMuted = False			# True if the "mute" button is toggled on
		self.actuallyIsMuted = False	# True if the instrument is muted (silent)
		self.isSolo = False				# True if the instrument is solo'd (only instrument active)
		self.isVisible = True			# True if the instrument should be displayed in the mixer views
		self.level = 0.0				# Current audio level in range 0..1
		self.volume = 0.5				# Gain of the current instrument in range 0..1
		self.instrType = type			# The type of instrument
		self.effects = []				# List of GStreamer effect elements
		self.pan = 0.0					# pan number (between -100 and 100)
		self.currentchainpreset = None	# current instrument wide chain preset
		
		# Select first input device as default to avoid a GStreamer bug which causes
		# large amounts of latency with the ALSA 'default' device.
		try:
			self.input = AlsaDevices.GetAlsaList("capture").values()[1]
		except: 
			self.input = "default"

		self.inTrack = 0	# Input track to record from
		self.output = ""
		self.recordingbin = None
		self.id = project.GenerateUniqueID(id) #check is id is already being used before setting
		self.isSelected = False			# True if the instrument is currently selected
		
		# GStreamer pipeline elements for this instrument		
		self.volumeElement = gst.element_factory_make("volume", "Instrument_Volume_%d"%self.id)
		self.levelElement = gst.element_factory_make("level", "Instrument_Level_%d"%self.id)
		self.levelElement.set_property("interval", gst.SECOND / 50)
		self.levelElement.set_property("message", True)
		self.levelElement.set_property("peak-ttl", 0)
		self.levelElement.set_property("peak-falloff", 20)

		self.panElement = gst.element_factory_make("audiopanorama", "Instrument_Pan_%d"%self.id)

		self.playbackbin = gst.element_factory_make("bin", "Instrument_%d"%self.id)

		self.playbackbin.add(self.volumeElement)
		Globals.debug("added volume (instrument)")

		self.playbackbin.add(self.levelElement)
		Globals.debug("added level (instrument)")

		self.playbackbin.add(self.panElement)
		Globals.debug("added audiopanorama (instrument)")

		# Create Effects Bin
		self.effectsBin = gst.element_factory_make("bin", "InstrumentEffects_%d"%self.id)
		self.playbackbin.add(self.effectsBin)
		# Create convert for start of effects bin
		self.effectsBinConvert = gst.element_factory_make("audioconvert", "Start_Effects_Converter_%d"%self.id)
		self.effectsBin.add(self.effectsBinConvert)
		# Create ghostpads for the bin from the audioconvert
		self.effectsBinSink = gst.GhostPad("sink", self.effectsBinConvert.get_pad("sink"))
		self.effectsBin.add_pad(self.effectsBinSink)
		self.effectsBinSrc = gst.GhostPad("src", self.effectsBinConvert.get_pad("src"))
		self.effectsBin.add_pad(self.effectsBinSrc)
		# Link the end of the effects bin to the level element
		self.effectsBin.link(self.volumeElement)
		
		
		self.composition = gst.element_factory_make("gnlcomposition")

		# adding default source - this adds silence betweent the tracks
		self.silenceaudio = gst.element_factory_make("audiotestsrc")
		self.silenceaudio.set_property("wave", 4)	#4 is silence
		
		self.silencesource = gst.element_factory_make("gnlsource")
		self.silencesource.set_property("priority", 2 ** 32 - 1)
		self.silencesource.set_property("start", 0)
		self.silencesource.set_property("duration", 1000 * gst.SECOND)
		self.silencesource.set_property("media-start", 0)
		self.silencesource.set_property("media-duration", 1000 * gst.SECOND)
		self.silencesource.add(self.silenceaudio)
		self.composition.add(self.silencesource)

		self.playbackbin.add(self.composition)
		Globals.debug("added composition (instrument)")
		
		self.resample = gst.element_factory_make("audioresample")
		self.playbackbin.add(self.resample)
		Globals.debug("added audioresample (instrument)")

		# create operation

		self.volbin = gst.element_factory_make("bin", "volbin")

		self.opvol = gst.element_factory_make("volume", "opvol")
		self.volbin.add(self.opvol)

		self.vollac = gst.element_factory_make("audioconvert", "vollac")
		self.volbin.add(self.vollac)
		
		self.volrac = gst.element_factory_make("audioconvert", "volrac")
		self.volbin.add(self.volrac)

		self.vollac.link(self.opvol)
		self.opvol.link(self.volrac)

		volbinsink = gst.GhostPad("sink", self.vollac.get_pad("sink"))
		self.volbin.add_pad(volbinsink)
		volbinsrc = gst.GhostPad("src", self.volrac.get_pad("src"))
		self.volbin.add_pad(volbinsrc)

		self.op = gst.element_factory_make("gnloperation", "gnloperation")
		self.op.set_property("start", long(0) * gst.SECOND)
		self.op.set_property("duration", long(20) * gst.SECOND)
		self.op.set_property("priority", 1)

		self.op.add(self.volbin)

		self.composition.add(self.op)
		
		# set controller

		self.control = gst.Controller(self.opvol, "volume")
		self.control.set_interpolation_mode("volume", gst.INTERPOLATE_LINEAR)

		# link elements
		
		#self.converterElement.link(self.volumeElement)
		#print "linked instrument audioconvert to instrument volume"

		self.volumeElement.link(self.levelElement)
		Globals.debug("linked instrument volume to instrument level")

		self.levelElement.link(self.panElement)
		self.panElement.set_property("panorama", 0)
		Globals.debug("linked instrument level to instrument pan")


		self.panElement.link(self.resample)
		Globals.debug("linked instrument pan to instrument resample")
		
		self.playghostpad = gst.GhostPad("src", self.resample.get_pad("src"))
		self.playbackbin.add_pad(self.playghostpad)
		Globals.debug("created ghostpad for instrument playbackbin")
		
		self.AddAndLinkPlaybackbin()

		self.composition.connect("pad-added", self.__PadAddedCb)
		self.composition.connect("pad-removed", self.__PadRemovedCb)
		
		#mute this instrument if another one is solo
		self.OnMute()
		
	#_____________________________________________________________________
	
	def __PadAddedCb(self, element, pad):
		""" Links a new pad to the rest of the playbackbin when one is created by the composition.
		"""
		Globals.debug("NEW PAD on instrument %s" % self.name)
		convpad = self.effectsBin.get_compatible_pad(pad, pad.get_caps())
		pad.link(convpad)

	#_____________________________________________________________________

	def __PadRemovedCb(self, element, pad):
		""" Removes a new GStreamer pad from the specified instrument.
		"""
		Globals.debug("pad removed on instrument %s" % self.name)
		self.composition.set_state(gst.STATE_READY)

	#_____________________________________________________________________
	
	def __repr__(self):
		return "Instrument [%d] %s"%(self.id, self.name)
		
	#_____________________________________________________________________
		
	def StoreToXML(self, doc, parent, graveyard=False):
		if graveyard:
			ins = doc.createElement("DeadInstrument")
		else:
			ins = doc.createElement("Instrument")
		parent.appendChild(ins)
		ins.setAttribute("id", str(self.id))
		
		items = ["name", "isArmed", 
				  "isMuted", "isSolo", "input", "output",
				  "isSelected", "isVisible", "inTrack", "instrType", "pan"]
		
		params = doc.createElement("Parameters")
		ins.appendChild(params)
		
		StoreParametersToXML(self, doc, params, items)
		
		for effect in self.effects:
			globaleffect = doc.createElement("GlobalEffect")
			globaleffect.setAttribute("element", effect.get_factory().get_name())
			ins.appendChild(globaleffect)
		
			propsdict = {}
			for prop in gobject.list_properties(effect):
				if prop.flags & gobject.PARAM_WRITABLE:
					propsdict[prop.name] = effect.get_property(prop.name)
			
			StoreDictionaryToXML(doc, globaleffect, propsdict)
			
		for e in self.events:
			e.StoreToXML(doc, ins)
		for e in self.graveyard:
			e.StoreToXML(doc, ins, graveyard=True)
			
	#_____________________________________________________________________	
			
	def LoadFromXML(self, node):
		
		params = node.getElementsByTagName("Parameters")[0]
		
		LoadParametersFromXML(self, params)
		
		#figure out the instrument's path based on the location of the projectfile
		self.path = os.path.join(os.path.dirname(self.project.projectfile), "audio")
		
		globaleffect = node.getElementsByTagName("GlobalEffect")
		
		for effect in globaleffect:
			elementname = str(effect.getAttribute("element"))
			Globals.debug("Loading effect:", elementname)
			instance = gst.element_factory_make(elementname, "effect")
			
			propsdict = LoadDictionaryFromXML(effect)
			for key, value in propsdict.iteritems():
				instance.set_property(key, value)		
			self.effects.append(instance)
			
		for ev in node.getElementsByTagName("Event"):
			try:
				id = int(ev.getAttribute("id"))
			except ValueError:
				id = None
			e = Event(self, None, id)
			e.LoadFromXML(ev)
			self.events.append(e)
	
		for ev in node.getElementsByTagName("DeadEvent"):
			try:
				id = int(ev.getAttribute("id"))
			except ValueError:
				id = None
			e = Event(self, None, id)
			e.LoadFromXML(ev)
			self.graveyard.append(e)
			#remove it from the composition so it doesnt play
			self.composition.remove(e.filesrc)
		
		#load image from file based on unique type
		#TODO replace this with proper cache manager
		for i in Globals.getCachedInstruments():
			if self.instrType == i[1]:
				self.pixbuf = i[2]
				break
		if not self.pixbuf:
			Globals.debug("Error, could not load image:", self.instrType)
		
		# load pan level
		
		self.panElement.set_property("panorama", self.pan)
		
		#check if instrument is muted and setup accordingly
		self.OnMute()

	#_____________________________________________________________________

	def AddEffect(self, effectName):
		"""
		Add an effect with the Gstreamer element name effectName
		to the pipeline for this instrument. The effect is always placed
		in the pipeline after any other effects that were previously added.
		"""
		#make the new effect and an audioconvert to go with it
		convert = gst.element_factory_make("audioconvert")
		effectElement = gst.element_factory_make(effectName)
		self.effects.append(effectElement)
		#add both elements to effects bin
		self.effectsBin.add(convert)
		self.effectsBin.add(effectElement)
		
		# The sink pad on the first element to the right of the bin
		externalSinkPad = self.effectsBinSrc.get_peer()
		# The src pad on the last element in the bin
		endSrcPad = self.effectsBinSrc.get_target()
		
		state = self.playbackbin.get_state(0)[1]
		if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING:
			endSrcPad.set_blocked(True)
		# Unlink the bin from the external element so we can put in a new ghostpad
		self.effectsBinSrc.unlink(externalSinkPad)
		# Remove the old ghostpad
		self.effectsBin.remove_pad(self.effectsBinSrc)
		
		newEffectPad = effectElement.sink_pads().next()
		# Link the last element in the bin with the new effect
		endSrcPad.link(newEffectPad)
		# Link the new element to the new convert
		effectElement.link(convert)
		# Make the audioconvert the new ghostpad
		self.effectsBinSrc = gst.GhostPad("src", convert.get_pad("src"))
		self.effectsBin.add_pad(self.effectsBinSrc)
		self.effectsBinSrc.link(externalSinkPad)
		
		# make the elements' state match the bin's state
		convert.set_state(state)
		effectElement.set_state(state)
		#give it a lambda for a callback that does nothing, so we don't have to wait
		endSrcPad.set_blocked_async(False, lambda x,y: False)
		
		self.StateChanged("effects")
		
		return effectElement
	
	#_____________________________________________________________________
	
	def RemoveEffect(self, effect):
		"""
		Remove the given Gstreamer element from the effects bin.
		"""
		if effect not in self.effects:
			Globals.debug("Error: trying to remove an element that is not in the list")
			return
		
		previousConvert = None
		for pad in effect.sink_pads():
			if pad.is_linked():
				previousConvert = pad.get_peer().get_parent()
				break
					
		nextConvert = None
		for pad in effect.src_pads():
			if pad.is_linked():
				nextConvert = pad.get_peer().get_parent()
				break
		
		state = self.playbackbin.get_state(0)[1]
		if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING:
			previousConvert.get_pad("src").set_blocked(True)
##			nextConvert.get_pad("src").get_peer().set_blocked(True)
		
		# If we have to remove from the end
		if self.effects[-1] == effect:
			# The sink pad on the first element to the right of the bin
			externalSinkPad = self.effectsBinSrc.get_peer()
			
			# Unlink the bin from the external element so we can put in a new ghostpad
			self.effectsBinSrc.unlink(externalSinkPad)
			# Remove the old ghostpad
			self.effectsBin.remove_pad(self.effectsBinSrc)
			
			previousConvert.unlink(effect)
			
			# Make the audioconvert the new ghostpad
			self.effectsBinSrc = gst.GhostPad("src", previousConvert.get_pad("src"))
			self.effectsBin.add_pad(self.effectsBinSrc)
			self.effectsBinSrc.link(externalSinkPad)
			
		# Else we are removing from the middle or beginning of the list
		else: 
			nextEffect = nextConvert.get_pad("src").get_peer().get_parent()
			nextConvert.unlink(nextEffect)
			
			previousConvert.unlink(effect)
			previousConvert.link(nextEffect)
			
		# Remove and dispose of the two elements
		effect.unlink(nextConvert)
		self.effectsBin.remove(effect)
		self.effectsBin.remove(nextConvert)
		effect.set_state(gst.STATE_NULL)
		nextConvert.set_state(gst.STATE_NULL)
		#remove the effect from our own list
		self.effects.remove(effect)
		
		#give it a lambda for a callback that does nothing, so we don't have to wait
		previousConvert.get_pad("src").set_blocked_async(False, lambda x,y: False)
		
		self.StateChanged("effects")
	
	#_____________________________________________________________________
	
	def ChangeEffectOrder(self, effect, newPosition):
		"""
		Move the given Gstreamer element from its current position
		in the effects bin, to the index newPosition.
		"""
		pass
	
	#_____________________________________________________________________
	
	def getRecordingEvent(self):
		event = Event(self)
		event.start = 0
		event.isRecording = True
		event.name = _("Recorded audio")
		event.file = "%s_%d_%d.ogg"%(os.path.join(self.path, self.name.replace(" ", "_")), self.id, int(time.time()))
		self.events.append(event)
		return event

	#_____________________________________________________________________

	@UndoCommand("DeleteEvent", "temp")
	def finishRecordingEvent(self, event):
		event.isRecording = False
		event.GenerateWaveform()
		self.temp = event.id
		self.StateChanged()
	
	#_____________________________________________________________________

	@UndoCommand("DeleteEvent", "temp")
	def addEventFromFile(self, start, file, copyfile=False):
		''' Adds an event to this instrument, and attaches the specified
			file to it. 
			
			start - The offset time in seconds
			file - file path
			copyfile - if True copy file to project audio dir
		'''
		filelabel=file

		if copyfile:
			basename = os.path.split(file.replace(" ", "_"))[1]
			basecomp = basename.rsplit('.', 1)
			if len(basecomp) > 1:
				newfile = "%s_%d_%d.%s" % (basecomp[0], self.id, int(time.time()), basecomp[len(basecomp)-1])
			else:
				newfile = "%s_%d_%d" % (basecomp[0], self.id, int(time.time()))

			audio_file = os.path.join(self.path, newfile)
			shutil.copyfile(file,audio_file)
			self.project.deleteOnCloseAudioFiles.append(audio_file)
			
			file = audio_file
			name = basename
		else:
			name = file.split(os.sep)[-1]

		e = Event(self, file,None,filelabel)
		e.start = start
		e.name = name
		self.events.append(e)
		e.GenerateWaveform()

		self.temp = e.id
		
		self.StateChanged()
		
		return e
		
	#_____________________________________________________________________
	
	@UndoCommand("DeleteEvent", "temp")
	def addEventFromURL(self, start, url):
		''' Adds an event to this instrument, and downloads the specified
			URL and saves it against this event.
			
			start - The offset time in seconds
			file - file path
			
			NB: there is no copyfile option here, because it's mandatory.
		'''
		
		# no way of knowing whether there's a filename, so make one up
		newfile = "%d_%d" % (self.id, int(time.time()))

		audio_file = os.path.join(self.path, newfile)
		self.project.deleteOnCloseAudioFiles.append(audio_file)
		
		# Create the event now so we can return it, and fill in the file later
		e = Event(self, None, None)
		e.start = start
		self.events.append(e)
		self.temp = e.id

		PRIORITY_DEFAULT = 0 # not wrapped in gnomevfs module yet
		gnomevfs.async.open(url, self.__got_url_handle, gnomevfs.OPEN_READ, 
		  PRIORITY_DEFAULT, [audio_file, start, e])

		return e
	
	#_____________________________________________________________________
	
	def __got_url_handle(self, handle, param, callbackdata):
		"""Called once gnomevfs has an object that we can read data from"""
		audio_file, start, event = callbackdata
		fp = open(audio_file, 'wb')
		handle.read(1024, self.__async_read_callback, [fp, start, event])
	
	#_____________________________________________________________________
	
	def __async_read_callback(self, handle, data, iserror, length, callbackdata):
		fp, start, event = callbackdata
		fp.write(data)
		if iserror is None:
			handle.read(1024, self.__async_read_callback, [fp, start, event])
		else:
			# all data now loaded
			Globals.debug("Event data downloaded")
			event.name = os.path.split(fp.name)[1]
			event.file = fp.name
			fp.close()
			Globals.debug("Creating filesource")
			event.CreateFilesource()
			Globals.debug("Generating waveform")
			event.GenerateWaveform()
			self.StateChanged()
		
	#_____________________________________________________________________
	
	@UndoCommand("DeleteEvent", "temp")
	def addEventFromEvent(self, start, event):
		"""Creates a new event instance identical to the event parameter 
		   and adds it to this instrument (for paste functionality).
		      start - The offset time in seconds
		      event - The event to be recreated on this instrument
		"""
		e = Event(self, event.file)
		e.start = start
		for i in ["duration", "name", "offset"]:
			setattr(e, i, getattr(event, i))
		e.levels = event.levels[:]
		
		self.events.append(e)
		e.SetProperties()
		e.MoveButDoNotOverlap(e.start)
		
		self.temp = e.id
		self.StateChanged()
	
	#_____________________________________________________________________
	
	def DeleteEvent(self, eventid):
		'''Removes an event from this instrument. 
		   It does not register with undo or append it to the graveyard,
		   because both are done by event.Delete()
		'''
		event = [x for x in self.events if x.id == eventid][0]
		event.Delete()
	
	#_____________________________________________________________________

	def MultipleEventsSelected(self):
		''' Confirms whether or not multiple events are selected '''
		multiple = 0
		for ev in self.events:
			if (ev.isSelected):
				multiple += 1
		return (multiple > 1)

	#_____________________________________________________________________

	def JoinEvents(self):
		''' Joins together all the selected events into a single event '''

		eventsToJoin = []
		for ev in self.events:
			if (ev.isSelected):
				eventsToJoin.append(ev)

		# Move them next to each other
		joinPoint = eventsToJoin[0].start + eventsToJoin[0].duration
		for ev in eventsToJoin[1:]:
			ev.Move(ev.start, joinPoint)
			joinPoint = joinPoint + ev.duration

		# Join them into a single event
		while (len(eventsToJoin) > 1):
			eventsToJoin[0].Join(eventsToJoin[1].id)
			eventsToJoin.remove(eventsToJoin[1])
			
	#_____________________________________________________________________

	def SetLevel(self, level):
		""" Note that this sets the current REPORTED level, NOT THE VOLUME!
		"""
		self.level = level
	
	#_____________________________________________________________________

	def SetVolume(self, volume):
		"""Sets the volume of the instrument in the range 0..1
		"""
		if self.volume != volume:
			self.volume = volume
			self.volumeElement.set_property("volume", volume)
			self.StateChanged("volume")

	#_____________________________________________________________________
	
	@UndoCommand("SetName", "temp")
	def SetName(self, name):
		"""
		   Sets the instrument to the given name
		   so it can be registered in the undo stack
		"""
		if self.name != name:
			self.temp = self.name
			self.name = name
			self.StateChanged()
	
	#_____________________________________________________________________
	
	@UndoCommand("ToggleArmed")
	def ToggleArmed(self):
		"""
		   Toggles the instrument to be armed for recording
		"""
		self.isArmed = not self.isArmed
		self.StateChanged()
		
	#_____________________________________________________________________
	
	@UndoCommand("ToggleMuted", "temp")
	def ToggleMuted(self, wasSolo):
		"""
		   Toggles the instrument to be muted
		"""
		self.temp = self.isSolo
		self.isMuted = not self.isMuted
		if self.isSolo and not wasSolo:
			self.isSolo = False
			self.project.soloInstrCount -= 1
			self.project.OnAllInstrumentsMute()
		elif not self.isSolo and wasSolo:
			self.isSolo = True
			self.project.soloInstrCount += 1
			self.project.OnAllInstrumentsMute()
		else:
			self.OnMute()
			
		self.StateChanged()
	
	#_____________________________________________________________________
	
	@UndoCommand("ToggleSolo", "temp")
	def ToggleSolo(self, wasMuted):
		"""
		   Toggles the all the other instruments muted
		"""
		self.temp = self.isMuted
		self.isMuted = wasMuted
		
		if self.isSolo:
			self.isSolo = False
			self.project.soloInstrCount -= 1
		else:
			self.isSolo = True
			self.project.soloInstrCount += 1
		
		self.project.OnAllInstrumentsMute()
		self.StateChanged()
	
	#_____________________________________________________________________
	
	@UndoCommand("SetVisible", "temp")
	def SetVisible(self, visible):
		"""
		   Sets wheather the instrument is minimized in CompactMixView
		"""
		if self.isVisible != visible:
			self.temp = self.isVisible
			self.isVisible = visible
			self.StateChanged()
	
	#_____________________________________________________________________
	
	def SetSelected(self, sel):
		"""
		   Sets the instrument to be highlighted 
		   and receive keyboard actions
		"""
		# No need to call StateChanged when there is no change in selection state
		if self.isSelected is not sel:
			self.isSelected = sel
			self.StateChanged()
	
	#_____________________________________________________________________
	
	def OnMute(self):
		self.checkActuallyIsMuted()
		if self.actuallyIsMuted:
			self.volumeElement.set_property("mute", True)
		else:
			self.volumeElement.set_property("mute", False)
	
	#_____________________________________________________________________
	
	def checkActuallyIsMuted(self):
		"""Determines if this intrument should be muted
		   by taking into account if any other intruments are muted
		"""
		if self.isMuted:
			self.actuallyIsMuted = True
		elif self.isSolo:
			self.actuallyIsMuted = False
		elif self.project.soloInstrCount > 0:
			self.actuallyIsMuted = True
		else:
			self.actuallyIsMuted = False
	
	#_____________________________________________________________________
	
	def AddAndLinkPlaybackbin(self):
		#make sure our playbackbin is in the same state so the pipeline can continue what it was doing
		status, state, pending = self.project.playbackbin.get_state(0)
		if pending != gst.STATE_VOID_PENDING:
			self.playbackbin.set_state(pending)
		else:
			self.playbackbin.set_state(state)
			
		if not self.playbackbin in list(self.project.playbackbin.elements()):
			self.project.playbackbin.add(self.playbackbin)
			Globals.debug("added instrument playbackbin to adder playbackbin", self.id)
		if not self.playghostpad.get_peer():
			self.playbackbin.link(self.project.adder)
			#give it a lambda for a callback that does nothing, so we don't have to wait
			self.playghostpad.set_blocked_async(False, lambda x,y: False)
			Globals.debug("linked instrument playbackbin to adder (project)")

	#_____________________________________________________________________
	
	def RemoveAndUnlinkPlaybackbin(self):
		#get reference to pad before removing self.playbackbin from project.playbackbin!
		pad = self.playghostpad.get_peer()
		
		if pad:
			status, state, pending = self.playbackbin.get_state(0)
			if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING or \
					pending == gst.STATE_PAUSED or pending == gst.STATE_PLAYING:
				self.playghostpad.set_blocked(True)
			self.playbackbin.unlink(self.project.adder)
			self.project.adder.release_request_pad(pad)
			Globals.debug("unlinked instrument playbackbin from adder")
		
		if self.playbackbin in list(self.project.playbackbin.elements()):
			self.project.playbackbin.remove(self.playbackbin)
			Globals.debug("removed instrument playbackbin from project playbackbin")
	
	#_____________________________________________________________________

	@UndoCommand("ChangeType", "temp", "temp2")
	def ChangeType(self, type, name):
		"""
		   Changes the intruments type to the type specified.
		   The given type must be loaded in the instrument list
		   in Globals or the image will not be found.
		"""
		
		self.temp = self.instrType
		self.temp2 = self.name
		
		pixbufList = [x[2] for x in Globals.getCachedInstruments() if x[1] == type]
		if type != self.instrType and pixbufList:
			pixbuf = pixbufList[0]
		else:
			return

		oldname = self.name.lower().replace(" ","")
		#if the instrument name has not been modified by the user, we can replace it.
		if oldname == self.instrType:
			self.name = name

		self.instrType = type
		self.pixbuf = pixbuf

		self.StateChanged("image")

	#_____________________________________________________________________

	def PrepareController(self):
		"""Fills the gst.Controller for the instrument with its list of fade times."""
		
		Globals.debug("Preparing the controller")
		# set the length of the operation to be the full length of the project
		self.op.set_property("duration", self.project.GetProjectLength() * gst.SECOND)
		self.control.unset_all("volume")
		firstpoint = False
		for ev in self.events:
			for point in ev.audioFadePoints:
				if ev.start + point[0] == 0.0:
					firstpoint = True
				#FIXME: remove vol=0.99 when gst.Controller is fixed to accept many consecutive 1.0 values.
				if point[1] == 1.0:
					vol = 0.99
				else:
					vol = point[1]
				Globals.debug("FADE POINT: time(%.2f) vol(%.2f)" % (ev.start + point[0], vol))
				self.control.set("volume", (ev.start + point[0]) * gst.SECOND, vol)
		if not firstpoint:
			Globals.debug("Set extra zero fade point")
			self.control.set("volume", 0, 0.99)
	#_____________________________________________________________________
	
#=========================================================================	
